/**
 *#########################################################################
 *
 * A component of the Gatherer application, part of the Greenstone digital
 * library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: John Thompson, Greenstone Digital Library, University of Waikato
 *
 * Copyright (C) 1999 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *########################################################################
 */
package org.greenstone.gatherer;

//import com.l2fprod.gui.*;
//import com.l2fprod.gui.plaf.skin.*;
//import com.l2fprod.util.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import javax.swing.plaf.*;
import javax.swing.text.*;
import org.greenstone.gatherer.Configuration;
import org.greenstone.gatherer.GAuthenticator;
import org.greenstone.gatherer.collection.CollectionManager;
import org.greenstone.gatherer.feedback.ActionRecorderDialog;
import org.greenstone.gatherer.file.FileManager;
import org.greenstone.gatherer.file.FileAssociationManager;
import org.greenstone.gatherer.gui.GUIManager;
import org.greenstone.gatherer.gui.ModalDialog;
import org.greenstone.gatherer.gui.Splash;
import org.greenstone.gatherer.gui.URLField;
import org.greenstone.gatherer.gui.WarningDialog;
import org.greenstone.gatherer.util.GSDLSiteConfig;
import org.greenstone.gatherer.util.StaticStrings;
import org.greenstone.gatherer.util.Utility;
// import sun.misc.*;

/** Containing the top-level "core" for the Gatherer, this class is the
 * common core for the GLI application and applet. It first parses the
 * command line arguments, preparing to update the configuration as
 * required. Next it loads several important support classes such as the
 * Configuration and Dictionary. Finally it creates the other important
 * managers and sends them on their way.
 * @author John Thompson, Greenstone Digital Library, University of Waikato
 * @version 2.3
 */

// How to catch All Exceptions including those from the AWT Event thread.
// Step 1: register a handler class in your main()
//
// System.setProperty("sun.awt.exception.handler", "YourHandler");
//
// Step 2: implement your handler class.
//
// public class YourHandler {
//   public void handle(Throwable thrown) {
//     eat the exception without dumping to the console.
//   }
// }

public class Gatherer
{
    static public Hashtable authentications = new Hashtable();

    static final private String SKIN_DEFINITION_FILE = "lib/greenaqua/greenaqua.xml";

    static public boolean always_show_exceptions = true;
    /** Has the exit flag been set? <i>true</i> if so, <i>false</i> otherwise. */
    public boolean exit = false;
    /** The size of the Gatherer window. */
    public Dimension frame_size = null;

    /** All of the external applications that must exit before we close the Gatherer. */
    public Vector apps = new Vector();
    /** A public reference to the FileAssociationManager. */
    static public FileAssociationManager assoc_man;
    /** A public reference to the CollectionManager. */
    static public CollectionManager c_man;
    /** A public reference to the Gatherer's configuration. */
    static public Configuration config;
    /** a reference to the Servlet Configuration is GS3 */
    static public ServletConfiguration servlet_config;
    /** The current modal dialog being shown on screen, if any. */
    static public ModalDialog current_modal = null;
    /** A public reference to the Dictionary. */
    static public Dictionary dictionary;
    /** A public reference to the FileManager. */
    static public FileManager f_man;
    /** A public reference to the GUIManager. */
    static public GUIManager g_man;
    static public boolean g_man_built = false;

    /** A static reference to ourselves. */
    static public Gatherer self;
    /** The debug print stream. */
    static public PrintStream debug;
    /** We are using the GLI for GS3 */
    static public boolean GS3 = false;

    static public boolean isGsdlRemote = false;
    static public String cgiBase = "";

    // feedback stuff
    /** is the feedback feature enabled? */
    static public boolean feedback_enabled = true;
    /** the action recorder dialog */
    static public ActionRecorderDialog feedback_dialog = null;
    /** Extra environment information which must be set before shell processes will run properly. Should always be null if the startup script/program has done its job properly. */
    static public String extra_env[] = null;
    private GSDLSiteConfig gsdlsite_cfg = null;
    private ExternalApplication server = null;

    /** Magic to allow Enter to fire the default button. */
    static {
      KeyStroke enter = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0);
      Keymap map = JTextComponent.getKeymap(JTextComponent.DEFAULT_KEYMAP);
      map.removeKeyStrokeBinding(enter);
    }

    public Gatherer() 
    {
	this.self = this;

	init_images();
    }

    protected void init_images()
    {
	Class base = this.getClass();
	Utility.BLANK_ICON = new ImageIcon(base.getResource("/images/blank.gif"));
	Utility.ERROR_ICON = new ImageIcon(base.getResource("/images/error.gif"));
	Utility.HELP_ICON  = new ImageIcon(base.getResource("/images/help.gif"));
	Utility.ON_ICON    = new ImageIcon(base.getResource("/images/check.gif"));
	Utility.OFF_ICON   = new ImageIcon(base.getResource("/images/cross.gif"));

	Utility.base = base;
    }


    public GUIManager init(Dimension size, String gsdl_path, String gsdl3_path, 
			   String exec_path, boolean debug_enabled, String perl_path, 
			   boolean no_load, String open_collection, 
			   String site_name, String servlet_path, 
			   boolean mirroring_enabled, String wget_version_str, 
			   String wget_path) 
    {

	// This will hopefully catch ctrl-c and terminate, and exit gracefully. However it is platform specific, and may not be supported by some JVMs.
	/** It does, but I get bloody sick of it working when the Gatherer hangs.
	    CTRLCHandler handler = new CTRLCHandler();
	    Signal.handle(new Signal("INT"), handler);
	    Signal.handle(new Signal("TERM"), handler);
	    handler = null;
	*/
	if (gsdl3_path != null && !gsdl3_path.equals("")) {
	    this.GS3 = true;
	} else {
	    gsdl3_path = null;
	}
	// Create the debug stream only if required.
	if(debug_enabled) {
	    try {
		Calendar now = Calendar.getInstance();
		StringBuffer name = new StringBuffer("debug");
		name.append(now.get(Calendar.DATE));
		name.append("-");
		name.append(now.get(Calendar.MONTH));
		name.append("-");
		name.append(now.get(Calendar.YEAR));
		name.append(".txt");
		this.debug = new PrintStream(new FileOutputStream(name.toString()));
		Properties props = System.getProperties();
		props.list(debug);
	    }
	    catch(Exception error) {
		///ystem.err.println("Error in Gatherer.init(): " + error);
		error.printStackTrace();
		System.exit(1);
	    }
	}
	try {
	    // Load Config
	    loadConfig(gsdl_path, gsdl3_path, exec_path, perl_path, mirroring_enabled, site_name);

	    // the feedback dialog has been loaded with a default locale, 
	    // now set the user specified one
	    if (feedback_enabled && feedback_dialog != null) {
		feedback_dialog.setLocale(config.getLocale("general.locale", true));
	    }
	    
	    // Read Dictionary
	    dictionary = new Dictionary(config.getLocale("general.locale", true), config.getFont("general.font", true));


	    if (gsdl_path == null) {
		missingGSDL();
	    }

	    // If we were given a server run it if neccessary.
	    if (!GS3 && config.exec_file != null) {
		startServerEXE();
	    }

	    // Having loaded the configuration (necessary to determine if certain warnings have been disabled) and dictionary, we now check if the necessary path variables have been provided.
	    if (config.exec_file == null && config.exec_address == null) {
		if (config.exec_file == null) {
		    Gatherer.println("config.exec_file is null");
		}
		if (config.exec_address == null) {
		    Gatherer.println("config.exec_address is null");
		}
		missingEXEC();
	    }

	    if (Gatherer.isGsdlRemote) {
		Gatherer.println("Not checking for perl path/exe");
	    }
	    else {
		// Perl path is a little different as it is perfectly ok to
		// start the Gatherer without providing a perl path
		boolean found_perl = false;
		if (config.perl_path != null) {
		    // See if the file pointed to actually exists
		    File perl_file = new File(config.perl_path);
		    found_perl = perl_file.exists();
		    perl_file = null;
		}
		if (config.perl_path == null || !found_perl) {
		    // Run test to see if we can run perl as is.
		    PerlTest perl_test = new PerlTest();
		    if (perl_test.found()) {
			// If so replace the perl path with the system
			// default (or null for unix).
			config.perl_path = perl_test.toString();
			found_perl = true;
		    }
		}
		if (!found_perl) {
		    // Time for an error message.
		    missingPERL();
		}
	    }

	    // also check for wget
	    boolean mirror_workflow = config.get(StaticStrings.WORKFLOW_MIRROR, false);
	    if (mirror_workflow) {
		wget_version_str = StaticStrings.NO_WGET_STR;
		// has the user specified a path?
		if (wget_path != null && !wget_path.equals("")) {
		    File wget_file = new File(wget_path);
		    if (wget_file.exists()) {
			// we assume its ok if its there
			wget_version_str = StaticStrings.WGET_STR;
		    } 
		}
		
		// it hasn't been set by the user so we use the greenstone one 
		if (wget_version_str.equals(StaticStrings.NO_WGET_STR)) {
		    // TODO fix for gs3
		    wget_path = Utility.getWGetPath(gsdl_path);
		    File wget_file = new File(wget_path);
		    if (!wget_file.exists()) {
			// we give up trying to find one
			missingWGET();
		    } else {
			// we have found one
			wget_version_str = StaticStrings.WGET_STR;
		    }
		} 
		
		if (wget_version_str.equals(StaticStrings.WGET_STR)) {
		    // we have found one, should we check the version??
		    wget_version_str = testWGetVersion(wget_path);
		    if (wget_version_str.equals(StaticStrings.WGET_OLD_STR)) {
			oldWGET();
		    }
		}
		
		// tell the config the new values
		config.setWGetPath(wget_path);
		config.setWGetVersion(wget_version_str);
	    }

	    // Set default font
	    setUIFont(config.getFont("general.font", true), config.getFont("general.tooltip_font", true));
	    // Set up proxy
	    setProxy();
	    // Now we set up an Authenticator
	    Authenticator.setDefault(new GAuthenticator());

	    assoc_man = new FileAssociationManager();
	    // Create File Manager
	    f_man = new FileManager();
	    // Create Collection Manager
	    c_man = new CollectionManager();
	    
	    if (GS3) {
		if (site_name==null) {
		    site_name = config.site_name;
		    servlet_path = null; // need to reset this
		}
		if (servlet_path == null) {
		    servlet_path = config.getServletPath();
		}
	    }

	    // If there was an open collection last session, reopen it.
	    if (open_collection == null) {
		open_collection = config.getString("general.open_collection", true);
	    }
	    if (!no_load && open_collection.length() > 0) {
		c_man.loadCollection(open_collection);
	    }

	}
	catch (Exception error) {
	    error.printStackTrace();
	}

	// Create GUI Manager (last) or else suffer the death of a thousand NPE's
	g_man = new GUIManager(size);

	return g_man;
    }



    public void run(Dimension size, Splash splash, GUIManager g_man) 
    {
	// Size and place the frame on the screen
	Rectangle bounds = config.getBounds("general.bounds", true);
	if (bounds == null) {
	    // Choose a sensible default value
	    bounds = new Rectangle(0, 0, 640, 480);
	}
	
	// Ensure width and height are reasonable
	size = bounds.getSize();
	if (size.width < 640) {
	    size.width = 640;
	}
	else if (size.width > config.screen_size.width) {
	    size.width = config.screen_size.width;
	}
	if (size.height < 480) {
	    size.height = 480;
	}
	else if (size.height > config.screen_size.height) {
	    size.height = config.screen_size.height;
	}
	
	if (splash != null) { splash.toFront(); }

	if (!g_man_built) {

	    g_man.display();

	    // Place the window in the desired location on the screen, if this is do-able (not under most linux window managers apparently. In fact you're lucky if they listen to any of your screen size requests).
	    g_man.setLocation(bounds.x, bounds.y);
	    g_man.setVisible(true);

	    // After the window has been made visible, check that it is in the correct place
	    // sometimes java places a window not in the correct place, 
	    // but with an offset. If so, we work out what the offset is
	    // and change the desired location to take that into account 
	    Point location = g_man.getLocation();
	    int x_offset = bounds.x - location.x;
	    int y_offset = bounds.y - location.y;
	    // If not, offset the window to move it into the correct location
	    if (x_offset > 0 || y_offset > 0) {
		///ystem.err.println("changing the location to "+(bounds.x + x_offset)+" "+ (bounds.y + y_offset));
		g_man.setLocation(bounds.x + x_offset, bounds.y + y_offset);
	    }

	    // The 'after-display' triggers several events which don't occur until after the visual components are actually available on screen. Examples of these would be the various html renderings, as they can't happen offscreen.
	    g_man.afterDisplay();
	    g_man_built = true;
	}
	else {
	    g_man.setVisible(true);
	}

	if (splash != null) {
	    // Hide the splash screen
	    splash.hide();
	    splash.destroy();
	    splash = null;
	}
    }


    /** Exits the Gatherer after ensuring that things needing saving are saved.
     * @see java.io.FileOutputStream
     * @see java.io.PrintStream
     * @see java.lang.Exception
     * @see javax.swing.JOptionPane
     * @see org.greenstone.gatherer.Configuration
     * @see org.greenstone.gatherer.collection.CollectionManager
     * @see org.greenstone.gatherer.gui.GUIManager
     */
    public void exit() {
	exit = true;
	// If we have an open collection make note of it.
	config.setString("general.open_collection", true, null);
	if(c_man.ready()) {
	    ///ystem.err.println("Collection open.");
	    if(c_man.saved()) {
		///ystem.err.println("Collection has been recently saved, so I'll remember it for next time.");
		config.setString("general.open_collection", true, c_man.getCollectionFilename());
	    }
	    c_man.closeCollection();
	}
	if(assoc_man != null) {
	    assoc_man.save();
	    assoc_man = null;
	}

	// Store the current position and size (if reasonable) of the Gatherer for next time
	Rectangle bounds = g_man.getBounds();
	config.setBounds("general.bounds", true, bounds);

	// Save configuration.
	saveConfig();

	// Get the gui to deallocate
	g_man.hide();
	g_man.destroy();
	g_man_built = false;


	// Flush dictionary
	// dictionary.destroy();

	// Flush debug
	if(debug != null) {
	    try {
		debug.flush();
		debug.close();
	    }
	    catch (Exception error) {
		error.printStackTrace();
	    }
	}

	// If we started a server, we should try to stop it.
	if(!GS3 && gsdlsite_cfg != null) {
	    stopServerEXE();
	}

	if(apps.size() == 0) {
	    if (!Gatherer.isGsdlRemote) {
		System.exit(0);
	    }
	}
	else {
	    JOptionPane.showMessageDialog(g_man, Dictionary.get("General.Outstanding_Processes"), Dictionary.get("General.Outstanding_Processes_Title"), JOptionPane.ERROR_MESSAGE);
	    g_man.hide();
	}
    }

    // Used to send messages to the local library
    // Warning: this has a lot of potential for nasty race conditions
    // The response code is returned immediately -- but this does not mean the local
    //   library action has finished!
    public void configServer(String command)
    {
	try {
	    String raw_url = config.exec_address.toString() + command;
	    URL url = new URL(raw_url);
	    Gatherer.println("Action: " + raw_url);
	    HttpURLConnection library_connection = (HttpURLConnection) url.openConnection();
	    int response_code = library_connection.getResponseCode();
	    if(HttpURLConnection.HTTP_OK <= response_code && response_code < HttpURLConnection.HTTP_MULT_CHOICE) {
		Gatherer.println("200 - Complete.");
	    }
	    else {
		Gatherer.println("404 - Failed.");
	    }
	    url = null;
	}
	catch (Exception ex) {
	    ex.printStackTrace();
	    Gatherer.printStackTrace(ex);
	}
    }

    // used to send reload coll messages to the tomcat server
    public void configGS3Server(String site, String command) {

	try {
	    // need to add the servlet name to the exec address
	    String raw_url = config.exec_address.toString() + config.getServletPath() + command;
	    URL url = new URL(raw_url);
	    Gatherer.println("Action: " + raw_url);
	    HttpURLConnection library_connection = (HttpURLConnection) url.openConnection();
	    int response_code = library_connection.getResponseCode();
	    if(HttpURLConnection.HTTP_OK <= response_code && response_code < HttpURLConnection.HTTP_MULT_CHOICE) {
		Gatherer.println("200 - Complete.");
	    }
	    else {
		Gatherer.println("404 - Failed.");
	    }
	    url = null;
	}
	catch(Exception exception) {
	    Gatherer.printStackTrace(exception);
	    ///ystem.err.println("Bad URL.");
	}
    }

    /** Retrieve the metadata directory, as required by any MSMCaller implementation.
     * @return The currently active collection metadata directory as a <strong>String</strong>.
     * @see org.greenstone.gatherer.collection.CollectionManager
     */
    public String getCollectionMetadata() {
	if (c_man != null && c_man.ready()) {
	    return c_man.getCollectionMetadata();
	}
	return "";
    }

    /** Used to 'spawn' a new child application when a file is double clicked.
     * @param file The file to open 
     * @see org.greenstone.gatherer.Gatherer.ExternalApplication
     */
    public void spawnApplication(File file) {
	String [] commands = assoc_man.getCommand(file);
	if(commands != null) {
	    ExternalApplication app = new ExternalApplication(commands);
	    apps.add(app);
	    app.start();
	}
	else {
	    ///ystem.err.println("No open command available.");
	}
    }
    
    /** Used to 'spawn' a new  browser application or reset an existing one when the preview button is clicked 
     * @param url The url to open the browser at
     * @see org.greenstone.gatherer.Gatherer.BrowserApplication
     */    
    public void spawnBrowser(String url) {
	String command = assoc_man.getBrowserCommand(url);
	if (command != null) {
	    BrowserApplication app = new BrowserApplication(command, url);
	    apps.add(app);
	    app.start();
	}
	else {
	    ///ystem.err.println("No browser command available."); 
	}
    }

    /** Prints a warning message about a missing library path, which means the final collection cannot be previewed in the Gatherer.
     */
    static public void missingEXEC() {
	String message_name;
	if (GS3) {
	    message_name = "warning.MissingEXEC_GS3";
	} else {
	    message_name = "warning.MissingEXEC";
	}
	WarningDialog dialog = new WarningDialog(message_name, "general.exec_address" );
	dialog.setValueField(new URLField(Gatherer.config.getColor("coloring.editable_foreground", false), Gatherer.config.getColor("coloring.editable_background", false), Gatherer.config.getColor("coloring.error_foreground", false), Gatherer.config.getColor("coloring.error_background", false)));
	dialog.display();
	dialog.dispose();
	dialog = null;

	String library_path_string = Gatherer.config.getString("general.exec_address", true);
	if (!library_path_string.equals("")) {
	    try {
		Gatherer.config.exec_address = new URL(library_path_string);
	    }
	    catch (MalformedURLException error) {
		///ystem.err.println("Error: Bad address: " + exec_address_string);
	    }
	}
   }

    /** Prints a warning message about a missing GSDL path, which although not fatal pretty much ensures nothing will work properly in the Gatherer.
     */
    static public void missingGSDL() {
	WarningDialog dialog = new WarningDialog("warning.MissingGSDL", false);
	dialog.display();
	dialog.dispose();
	dialog = null;
   }

    /** Prints a warning message about a missing PERL path, which although not fatal pretty much ensures no collection creation/building will work properly in the Gatherer. */
    static public void missingPERL() {
	WarningDialog dialog = new WarningDialog("warning.MissingPERL", false);
	dialog.display();
	dialog.dispose();
	dialog = null;
    }
    
    /** Prints a warning message about a missing a valid WGet path. not fatal, but mirroring won't work */
    static public void missingWGET() {
	WarningDialog dialog = new WarningDialog("warning.MissingWGET", false);
	dialog.display();
	dialog.dispose();
	dialog = null;
    }
    /** Prints a warning message about having an old version of WGet. not fatal, but mirroring may not work properly */
    static public void oldWGET() {
	WarningDialog dialog = new WarningDialog("warning.OldWGET", false);
	dialog.display();
	dialog.dispose();
	dialog = null;
    }
    /** Print a message to the debug stream. */
    static synchronized public void print(String message) {
	if(debug != null) {
	    debug.print(message);
	    System.err.print(message);
	}
    }
    /** Print a message to the debug stream. */
    static synchronized public void println(String message) {
	if(debug != null) {
	    debug.println(message);
	    System.err.println(message);
	}
    }

    /** Print a stack trace to the debug stream. */
    static synchronized public void printStackTrace(Exception exception) {
	if(debug != null) {
	    exception.printStackTrace(debug);
	    exception.printStackTrace();
	}
    }
    /** Prints a usage message to screen.
     */
    static public void printUsage(Dictionary dictionary) {
	System.out.println(Dictionary.get("General.Usage"));
    }

    /** Sets up the proxy connection by setting JVM Environment flags and creating a new Authenticator.
     * @see java.lang.Exception
     * @see java.lang.System
     * @see java.net.Authenticator
     * @see org.greenstone.gatherer.Configuration
     * @see org.greenstone.gatherer.GAuthenticator
     */
    static public void setProxy() {
	try {// Can throw several exceptions
	    if(Gatherer.config.get("general.use_proxy", true)) {
		System.setProperty("http.proxyType", "4");
		System.setProperty("http.proxyHost", Gatherer.config.getString("general.proxy_host", true));
		System.setProperty("http.proxyPort", Gatherer.config.getString("general.proxy_port", true));
		System.setProperty("http.proxySet", "true");
	    } else {
		System.setProperty("http.proxySet", "false");
	    }
	} catch (Exception error) {
	    Gatherer.println("Error in Gatherer.initProxy(): " + error);
	    Gatherer.printStackTrace(error);
	}
    }

    /** Set all the default fonts of the program to a choosen font, except the tooltip font which should be some fixed width font.
     * @param f The default font to use in the Gatherer as a <strong>FontUIResource</strong>.
     * @param ttf The tooltip font to use also as a <strong>FontUIResource</strong>.
     * @see java.util.Enumeration
     * @see javax.swing.UIManager
     */
    static private void setUIFont(FontUIResource f, FontUIResource ttf) {
	// sets the default font for all Swing components.
	// ex.
	//  setUIFont (new FontUIResource("Serif",Font.ITALIC,12));
	//
	Enumeration keys = UIManager.getDefaults().keys();
	while (keys.hasMoreElements()) {
	    Object key = keys.nextElement();
	    Object value = UIManager.get (key);
	    if (value instanceof FontUIResource)
		UIManager.put (key, f);
	}
	// Now set the tooltip font to some fixed width font
	UIManager.put("ToolTip.font", ttf);
    }

    /** Loads the configuration file if one exists. Otherwise it creates a new one. Currently uses serialization.
     * @param gsdl_path The path to the gsdl directory, gathered from the startup arguments, and presented as a <strong>String</strong>.
     * @param exec_path The path to the library executable, gathered from the startup arguments, and presented as a <strong>String</strong>.
     * @param perl_path The path to the PERL compiler as a <strong>String</strong>. Necessary for windows platform versions.
     * @param mirroring_enabled
     * @param wget_version_str
     * @param wget_path
     * @see java.io.FileInputStream
     * @see java.io.ObjectInputStream
     * @see java.lang.Exception
     * @see org.greenstone.gatherer.Configuration
     */
    private void loadConfig(String gsdl_path, String gsdl3_path, String exec_path, String perl_path, boolean mirroring_enabled, String site_name) {
	try {
	    config = new Configuration(gsdl_path, gsdl3_path, exec_path, perl_path, mirroring_enabled, site_name);
	}
	catch (Exception error) {
	    Gatherer.println(Configuration.CONFIG_XML+" is not a well formed XML document.");
	    Gatherer.printStackTrace(error);
	}
	if (GS3) {
	    try {
		servlet_config = new ServletConfiguration(gsdl3_path);
	    } catch (Exception error) {
		error.printStackTrace();
	    }
	}
    }

    /** Causes the general configuration file to export itself to xml. Doesn't effect any remaining collection configuration, as its up to the collection manager to handle them (especially since we have no idea where they are going). */
    private void saveConfig() {
	try {
	    config.save();
	} catch (Exception error) {
	    Gatherer.printStackTrace(error);
	}
    }

    private void startServerEXE() {
	if(config.exec_file != null && config.exec_address == null && Utility.isWindows() && !GS3) {
	    // First of all we create a GSDLSiteCFG object and check if a URL is already present, in which case the server is already running.
	    gsdlsite_cfg = new GSDLSiteConfig(config.exec_file);
	    String url = gsdlsite_cfg.getURL();
	    // If its already running then set exec address.
	    if(url != null) {
		try {
		    config.exec_address = new URL(url);
		}
		catch(Exception error) {
		}
	    }
	    // Otherwise its time to run the server in a spawned process.
	    if(config.exec_address == null && config.exec_file.exists()) {
		// Configure for immediate entry. Note that this only works if the gsdlsite.cfg file exists.
		gsdlsite_cfg.set();
		// Spawn server
		String command = config.exec_file.getAbsolutePath() + " " + gsdlsite_cfg.getSiteConfigFilename();
		server = new ExternalApplication(command);
		server.start();
		command = null;
		// Now we have to wait until program has started. We do this by reloading and checking
		///ystem.err.print("Waiting until the local library has loaded");
		try {
		    gsdlsite_cfg.load();
		    
		    int try_again = JOptionPane.YES_OPTION;
		    int attempt_count = 0;
		    while(gsdlsite_cfg.getURL() == null && try_again == JOptionPane.YES_OPTION) {
			///ystem.err.print(".");
			if(attempt_count == 60) {
			    attempt_count = 0;
			    try_again = JOptionPane.showConfirmDialog(Gatherer.g_man, Dictionary.get("Server.QuitTimeOut"), Dictionary.get("General.Warning"), JOptionPane.YES_NO_OPTION);
			}
			else {
			    synchronized(this) {
				wait(1000); // Wait one second (give or take)
			    }
			    gsdlsite_cfg.load();
			    attempt_count++;
			}
		    }
		    
		    if((url = gsdlsite_cfg.getURL()) != null) {
			// Ta-da. Now the url should be available.
			config.exec_address = new URL(url);
			
			// A quick test involves opening a connection to get the home page 	for this collection. If this fails then we try changing the url to be localhost.
			try {
			    Gatherer.println("Try connecting to server on config url: '" + config.exec_address + "'");
			    URLConnection connection = config.exec_address.openConnection();
			    connection.getContent();
			}
			catch(IOException bad_url_connection) {
			    try {
				Gatherer.println("Try connecting to server on local host: '" + gsdlsite_cfg.getLocalHostURL() + "'");
				config.exec_address = new URL(gsdlsite_cfg.getLocalHostURL	());
				URLConnection connection = config.exec_address.openConnection();
				connection.getContent();
			    }
			    catch(IOException worse_url_connection) {
				Gatherer.println("Can't connect to server on either address.");
				config.exec_address = null;
				config.exec_file = null;
			    }
			}
		    }
		    // Unable to start local library. Show appropriate message.
		    else {
			missingEXEC();
		    }
		}
		catch (Exception error) {
		    error.printStackTrace();
		}
	    }
	    // Can't do a damb thing.
	}
	Gatherer.println("Having started server.exe, exec_address is: " + config.exec_address);
    }

    private void stopServerEXE() {
	if(server != null && config.exec_address != null) {
	    // See if its already exited for some reason.
	    gsdlsite_cfg.load();
	    if(gsdlsite_cfg.getURL() != null) {
		// Send the command for it to exit.
		configServer(GSDLSiteConfig.QUIT_COMMAND);
		// Wait until it exits.
		try {
		    gsdlsite_cfg.load();
		    int try_again = JOptionPane.YES_OPTION;
		    int attempt_count = 0;
		    while(gsdlsite_cfg.getURL() != null && try_again == JOptionPane.YES_OPTION) {
			if(attempt_count == 60) {
			    attempt_count = 0;
			    try_again = JOptionPane.showConfirmDialog(Gatherer.g_man, Dictionary.get("Server.QuitTimeOut"), Dictionary.get("General.Warning"), JOptionPane.YES_NO_OPTION);
			}
			else {
			    synchronized(this) {
				wait(1000); // Wait one second (give or take)
			    }
			    gsdlsite_cfg.load();
			    attempt_count++;
			}
		    }
		    //if(gsdlsite_cfg.getURL() != null) {
		    //JOptionPane.showMessageDialog(Gatherer.g_man, Dictionary.get("Server.QuitManual"), Dictionary.get("General.Error"), JOptionPane.ERROR_MESSAGE);
		    //}
		}
		catch (Exception error) {
		    error.printStackTrace();
		}
	    }
	    // Restore the gsdlsite_cfg.
	    if(gsdlsite_cfg != null) {
		gsdlsite_cfg.restore();
	    }
	    // If the local server is still running then our changed values will get overwritten.
	    if(gsdlsite_cfg.getURL() != null) {
		JOptionPane.showMessageDialog(Gatherer.g_man, Dictionary.get("Server.QuitFailed"), Dictionary.get("General.Error"), JOptionPane.ERROR_MESSAGE);
	    }
	    gsdlsite_cfg = null;
	    server = null;
	}
    }

    // TODO fill this in
    private String testWGetVersion(String wget_path) {
	return StaticStrings.WGET_STR;
    }
    
    /** This private class contains an instance of an external application running within a JVM shell. It is important that this process sits in its own thread, but its more important that when we exit the Gatherer we don't actually System.exit(0) the Gatherer object until the user has volunteerily ended all of these child processes. Otherwise when we quit the Gatherer any changes the users may have made in external programs will be lost and the child processes are automatically deallocated. */
    private class ExternalApplication
	extends Thread {
	private Process process = null;
	/** The initial command string given to this sub-process. */
	private String command = null;
	private String[] commands = null;
	/** Constructor.
	 * @param command The initial command <strong>String</strong>.
	 */
	public ExternalApplication(String command) {
	    this.command = command;
	}

	public ExternalApplication(String[] commands) {
	    this.commands = commands;
	}
	/** We start the child process inside a new thread so it doesn't block the rest of Gatherer.
	 * @see java.lang.Exception
	 * @see java.lang.Process
	 * @see java.lang.Runtime
	 * @see java.lang.System
	 * @see java.util.Vector
	 */
	public void run() {
	    // Call an external process using the args.
	    try {
		if(commands != null) {
		    StringBuffer whole_command = new StringBuffer();
		    for(int i = 0; i < commands.length; i++) {
			whole_command.append(commands[i]);
			whole_command.append(" ");
		    }
		    println("Running " + whole_command.toString());
		    Runtime rt = Runtime.getRuntime();
		    process = rt.exec(commands);
		    process.waitFor();
		}
		else {
		    println("Running " + command);
		    Runtime rt = Runtime.getRuntime();
		    process = rt.exec(command);
		    process.waitFor();
		}
	    }
	    catch (Exception error) {
		println("Error in ExternalApplication.run(): " + error);
		printStackTrace(error);
	    }
	    // Remove ourself from Gatherer list of threads.
	    apps.remove(this);
	    // Call exit if we were the last outstanding child process thread.
	    if(apps.size() == 0 && exit == true) {
		stopServerEXE();
		System.exit(0);
	    }
	}
	public void stopExternalApplication() {
	    if(process != null) {
		process.destroy();
	    }
	}
    }
    /** This private class contains an instance of an external application running within a JVM shell. It is important that this process sits in its own thread, but its more important that when we exit the Gatherer we don't actually System.exit(0) the Gatherer object until the user has volunteerily ended all of these child processes. Otherwise when we quit the Gatherer any changes the users may have made in external programs will be lost and the child processes are automatically deallocated. */
    private class BrowserApplication
	extends Thread {
	private Process process = null;
	/** The initial command string given to this sub-process. */
	private String command = null;
	private String url = null;
	private String[] commands = null;
	/** Constructor.
	 * @param command The initial command <strong>String</strong>.
	 */
//  	public BrowserApplication(String command) {
//  	    this.command = command;
//  	}

	public BrowserApplication(String command, String url) {
	    StringTokenizer st = new StringTokenizer(command);
	    int num_tokens = st.countTokens();
	    this.commands = new String [num_tokens];
	    int i=0;
	    while (st.hasMoreTokens()) {
		commands[i] = st.nextToken();
		i++;
	    }
	    //this.commands = commands;
	    this.url = url;
	}
	/** We start the child process inside a new thread so it doesn't block the rest of Gatherer.
	 * @see java.lang.Exception
	 * @see java.lang.Process
	 * @see java.lang.Runtime
	 * @see java.lang.System
	 * @see java.util.Vector
	 */
	public void run() {
	    // Call an external process using the args.
	    if(commands == null) {
		apps.remove(this);
		return;
	    }
	    try {
		String prog_name = commands[0];
		String lower_name = prog_name.toLowerCase();
		if (lower_name.indexOf("mozilla") != -1 || lower_name.indexOf("netscape") != -1) { 
		    Gatherer.println("found mozilla or netscape, trying remote it");
		    // mozilla and netscape, try using a remote command to get things in the same window
		    String [] new_commands = new String[] {prog_name, "-raise", "-remote", "openURL("+url+",new-tab)"};
		    if(debug != null) {
			printArray(new_commands);
		    }
		    Runtime rt = Runtime.getRuntime();
		    process = rt.exec(new_commands);
		    int exitCode = process.waitFor();
		    if (exitCode != 0) { // if Netscape or mozilla was not open
			Gatherer.println("couldn't do remote, trying original command");
			if(debug != null) {
			    printArray(commands);
			}
			process = rt.exec(commands); // try the original command
		    }
		} else {
		    // just run what we have been given
		    StringBuffer whole_command = new StringBuffer();
		    for(int i = 0; i < commands.length; i++) {
			whole_command.append(commands[i]);
			whole_command.append(" ");
		    }
		    println("Running " + whole_command.toString());
		    Runtime rt = Runtime.getRuntime();
		    process = rt.exec(commands);
		    process.waitFor();
		}
	    }
	    
	    catch (Exception error) {
		println("Error in BrowserApplication.run(): " + error);
		printStackTrace(error);
	    }
	    // Remove ourself from Gatherer list of threads.
	    apps.remove(this);
	    // Call exit if we were the last outstanding child process thread.
	    if(apps.size() == 0 && exit == true) {
		stopServerEXE();
		System.exit(0);
	    }
	}
	public void printArray(String [] array) {
	    for(int i = 0; i < array.length; i++) {
		debug.print(array[i]+" ");
		System.err.println(array[i]+" ");
	    }
	}
	public void stopBrowserApplication() {
	    if(process != null) {
		process.destroy();
	    }
	}
    }

    /** This class is intented to detect a specific SIGNAL, in this case SIGINT, and exit properly, rather than letting the Gatherer be interrupted which has the potential to leave erroneous lock files. */
//      private class CTRLCHandler
//  	implements SignalHandler {
//  	/** <i>true</i> if we ignore any other signals we receive, most likely because we are already dealing with one, <i>false</i> otherwise. */
//  	private boolean ignore = false;
//  	/** The method called by the system to inform us a signal has occured.
//  	 * @param sig The <strong>Signal</strong> itself.
//  	 * @see org.greenstone.gatherer.collection.CollectionManager
//  	 */
//  	public void handle(Signal sig) {
//  	    if(!ignore) {
//  		ignore = true;
//  		// handle SIGINT
//  		System.out.println("Caught Ctrl-C...");
//  		if(c_man != null && c_man.ready()) {
//  		    c_man.closeCollection();
//  		}
//  		exit();
//  		ignore = false;
//  	    }
//  	}
//      }

    private class PerlTest {

	private String[] command = new String[2];

	public PerlTest() {
	    if(Utility.isWindows()) {
		command[0] = Utility.PERL_EXECUTABLE_WINDOWS;
	    }
	    else {
		command[0] = Utility.PERL_EXECUTABLE_UNIX;
	    }
	    command[1] = "-version";
	}

	public boolean found() {
	    boolean found = false;
	    try {
		Process prcs = Runtime.getRuntime().exec(command);
		prcs.waitFor();
		found = (prcs.exitValue() == 0);
		prcs = null;
	    }
	    catch(Exception error) {
	    }
	    return found;
	}

	public String toString() {
	    return command[0];
	}
    }
}
